// Generated by CoffeeScript 1.3.3
var BaseQuery, EventEmitter, assert, fluidize, inspect, toRelation,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; };

inspect = require('util').inspect;

EventEmitter = require('events').EventEmitter;

toRelation = require('../nodes').toRelation;

assert = require('assert');

fluidize = require('../fluid');

module.exports = BaseQuery = (function(_super) {

  __extends(BaseQuery, _super);

  /*
    The base class for all queries. While this class itself is not part of
    gesundheits public API, the methods defined on it are.
  */


  BaseQuery.rootNode = null;

  function BaseQuery(engine, opts) {
    var table;
    if (opts == null) {
      opts = {};
    }
    /*
        :param engine: The engine the query will be bound to.
    
        :param opts.table:
          a ``String``, ``Relation``, ``Alias``, or object literal with a single
          key and value which will be interpreted as an alias name and table,
          respectively. This is normally given to as the first parameter to the
          query creation functions in :mod:`queries/index`
    */

    this.bind(engine);
    if (table = opts.table) {
      table = toRelation(table);
    }
    this.q = new this.constructor.rootNode(table);
  }

  BaseQuery.prototype.copy = function() {
    /* Instantiate a new query with a deep copy of this ones AST
    */

    var c;
    c = new this.constructor(this.engine);
    c.q = this.q.copy();
    c.bind(this.engine);
    return c;
  };

  BaseQuery.prototype.visit = function(fn) {
    /*
        Call the given function in the context of this query. This is useful with
        query factory functions where you can use it as a sort-of-DSL::
    
            SELECT('people', ['name'], function(q) {
              // this === q
              this.join('addresses', {
                on: {person_id: q.project('people', 'id')},
                fields: ['city', 'region']
              })
            })
    */
    return fn.call(this);
  };

  BaseQuery.prototype.bind = function(engine) {
    /*
        Bind this query object to a new engine.
        If no argument is given the query will be bound to the default engine.
    */

    var oldEngine, _base, _ref, _ref1;
    oldEngine = this.engine;
    this.engine = engine || require('../').defaultEngine;
    if (this.engine !== oldEngine) {
      if (oldEngine != null) {
        if (typeof oldEngine.unextendQuery === "function") {
          oldEngine.unextendQuery(this);
        }
      }
      if (typeof (_base = this.engine).extendQuery === "function") {
        _base.extendQuery(this);
      }
    }
    assert((_ref = this.engine) != null ? _ref.query : void 0, "Engine has no query method: " + (inspect(this.engine)));
    return assert((_ref1 = this.engine) != null ? _ref1.render : void 0, "Engine has no render method: " + (inspect(this.engine)));
  };

  BaseQuery.prototype.render = function() {
    /*
        Render the query to a SQL string.
    */
    return this.engine.render(this.q);
  };

  BaseQuery.prototype.compile = function() {
    /*
        Compile this query object, returning a SQL string and parameter array.
    */

    var ret;
    console.log(ret = [this.render(), this.q.params()]);
    return ret;
  };

  BaseQuery.prototype.execute = function(cb) {
    /*
        Execute the query using ``@engine`` and return a `QueryAdapter`.
    
        :param cb: An (optional) node-style callback that will be called with any
          errors and/or the query results. If no callback is given, a new EventEmitter
          will be returned that emits either an 'error' or 'result' event.
    */
    return this.engine.query.apply(this.engine, this.compile().concat(cb));
  };

  BaseQuery.prototype.toString = function() {
    return this.render();
  };

  return BaseQuery;

})(EventEmitter);

fluidize(BaseQuery, 'bind', 'visit');
