// Generated by CoffeeScript 1.3.3
var And, BaseQuery, CONST_NODES, Node, Or, OrderBy, SUDQuery, fluidize, nodes,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; },
  __slice = [].slice;

BaseQuery = require('./base');

nodes = require('../nodes');

fluidize = require('../fluid');

Node = nodes.Node, And = nodes.And, Or = nodes.Or, OrderBy = nodes.OrderBy, CONST_NODES = nodes.CONST_NODES;

module.exports = SUDQuery = (function(_super) {

  __extends(SUDQuery, _super);

  function SUDQuery() {
    return SUDQuery.__super__.constructor.apply(this, arguments);
  }

  /*
    SUDQuery is the base class for SELECT, UPDATE, and DELETE queries. It adds
    logic to :class:`queries/base::BaseQuery` for dealing with WHERE clauses and
    ordering.
  */


  SUDQuery.prototype.where = function(constraint) {
    /*
        Add a WHERE clause to the query.
    
        The where clause itself can be a comparison node, such as those produced by
        the :class:`nodes::ComparableMixin` methods::
    
          q.where(q.p('table','field1').eq(42))
          q.where(q.p('table','field2').gt(42))
    
        ... Or an object literal where each key is a field name (or field name
        alias) and each value is a constraint::
    
          q.where({field1: 42, field2: {gt: 42}})
    
        Constraint values can also be other projected fields::
    
          p = q.project.bind(q, 'table')
          q.where('table', p('field1').gt(p('field2')))
    
    
        To create a set of constraints joined by the OR operator, use 'or' as a key in
        the object literal with an array of further constraints. Similarly, you can use
        'and' as a key to nest 'and' constraints within an OR, nesting clauses arbitrarily
        deep.
    
          select('t').where({or: [{a: 1, and: [{b: 2, c: 3}]}]})
    
        Will generate the SQL statement::
    
          SELECT * FROM t WHERE (t.a = 1 OR (t.b = 2 AND t.c = 3))
    */

    var node, _i, _len, _ref, _results;
    _ref = this.makeClauses(constraint);
    _results = [];
    for (_i = 0, _len = _ref.length; _i < _len; _i++) {
      node = _ref[_i];
      _results.push(this.q.where.addNode(node));
    }
    return _results;
  };

  SUDQuery.prototype.makeClauses = function(constraint) {
    /*
        Return an array of Binary, And, and Or nodes for this constraint object
    */

    var clauses, column, field, item, op, predicate, val, _i, _len;
    clauses = [];
    if (Array.isArray(constraint)) {
      for (_i = 0, _len = constraint.length; _i < _len; _i++) {
        item = constraint[_i];
        if (item instanceof Node) {
          clauses.push(item);
        } else {
          clauses = clauses.concat(this.makeClauses(item));
        }
      }
      return clauses;
    }
    if (constraint instanceof Node) {
      return [constraint];
    }
    for (field in constraint) {
      predicate = constraint[field];
      if (field === 'and') {
        clauses.push(new And(this.makeClauses(predicate)));
      } else if (field === 'or') {
        debugger;
        clauses.push(new Or(this.makeClauses(predicate)));
      } else {
        column = this.project(field);
        if (predicate === null) {
          clauses.push(column.compare('IS', CONST_NODES.NULL));
        } else if (predicate.constructor === Object) {
          for (op in predicate) {
            val = predicate[op];
            clauses.push(column.compare(op, val));
          }
        } else {
          clauses.push(column.eq(predicate));
        }
      }
    }
    return clauses;
  };

  SUDQuery.prototype.or = function() {
    var clauses;
    clauses = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    /* Shortcut for ``.where({or: clauses})
    */

    return this.where({
      or: clauses
    });
  };

  SUDQuery.prototype.and = function() {
    var clauses;
    clauses = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    return this.where({
      and: clauses
    });
  };

  SUDQuery.prototype.order = function() {
    var args, dir, direction, field, name, orderBy, orderings, rel, _i, _j, _len, _len1, _ref, _results;
    args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
    /*
        Add an ORDER BY to the query. Currently this *always* uses the "active"
        table of the query. (See :meth:`queries/select::SelectQuery.from`)
    
        Each ordering can either be a string, in which case it must be a valid-ish
        SQL snippet like 'some_field DESC', (the field name and direction will still
        be normalized) or an object, in which case each key will be treated as a
        field and each value as a direction.
    */

    rel = this.defaultRel();
    orderings = [];
    for (_i = 0, _len = args.length; _i < _len; _i++) {
      orderBy = args[_i];
      switch (orderBy.constructor) {
        case String:
          orderings.push(orderBy.split(' '));
          break;
        case OrderBy:
          this.q.orderBy.addNode(orderBy);
          break;
        case Object:
          for (name in orderBy) {
            dir = orderBy[name];
            orderings.push([name, dir]);
          }
          break;
        default:
          throw new Error("Can't turn " + orderBy + " into an OrderBy object");
      }
    }
    _results = [];
    for (_j = 0, _len1 = orderings.length; _j < _len1; _j++) {
      _ref = orderings[_j], field = _ref[0], direction = _ref[1];
      direction = (function() {
        switch ((direction || '').toLowerCase()) {
          case 'asc':
          case 'ascending':
            return 'ASC';
          case 'desc':
          case 'descending':
            return 'DESC';
          case '':
            return '';
          default:
            throw new Error("Unsupported ordering direction " + direction);
        }
      })();
      _results.push(this.q.orderBy.addNode(new OrderBy(rel.project(field), direction)));
    }
    return _results;
  };

  SUDQuery.prototype.limit = function(l) {
    /* Set the LIMIT on this query
    */
    return this.q.limit.value = l;
  };

  SUDQuery.prototype.offset = function(l) {
    /* Set the OFFSET of this query
    */
    return this.q.offset.value = l;
  };

  SUDQuery.prototype.defaultRel = function() {
    return this.q.relations.active;
  };

  SUDQuery.prototype.project = function(relation, field) {
    /*
        Return a :class:`nodes::Projection` node representing ``<relation>.<field>``.
    
        The first argument is optional and specifies a table or alias name referring
        to a relation already joined to this query. If you don't specify a relation,
        the table added or focused last will be used. Alternatively, you can specify
        the relation name and field with a single dot-separated string::
    
          q.project('departments.name') == q.project('departments', 'name')
    
        The returned object has a methods from :class:`nodes::ComparableMixin` that
        can create new comparison nodes usable in join conditions and where clauses::
    
          # Find developers over the age of 45
          s = select('people', ['name'])
          s.join('departments', on: {id: s.project('people', 'department_id')})
          s.where(s.project('departments', 'name').eq('development'))
          s.where(s.project('people', 'age').gte(45))
    
        ``project`` is also aliased as ``p`` for those who value brevity::
    
             q.where(q.p('departments.name').eq('development'))
    
        .. note:: this means you *must* specify a relation name if you have a field
          name with a dot in it, if you have dots in your column names, sorry.
    */

    var rel, rel_field;
    if (field != null) {
      rel = this.q.relations.get(relation, true);
    } else {
      field = relation;
      rel_field = field.split('.');
      if (rel_field.length === 2) {
        field = rel_field[1];
        rel = this.q.relations.get(rel_field[0], true);
      } else {
        rel = this.defaultRel();
      }
    }
    return rel.project(field);
  };

  return SUDQuery;

})(BaseQuery);

fluidize(SUDQuery, 'where', 'or', 'and', 'limit', 'offset', 'order');

SUDQuery.prototype.p = SUDQuery.prototype.project;
